--================================
-- environmentHelper.ms v.1.0
-- Date: February 05, 2015
--
-- Brought to you by: the Autodesk Scripting and SDK Learning Channel - Autodesk M&E
--  Copyright 2015 Autodesk, Inc. All rights reserved.
-- Conceived & Designed by Amer Yassine - Autodesk M&E
-- Written & Developed by Martin Ashton - Autodesk M&E
-- [ youtube.com/3dsmaxhowtos | youtube.com/scriptingandsdkhowtos ]
-- [ 3dsmaxhowto@autodesk.com | scriptingandsdkhowto@autodesk.com ]
--
-------------------------
-- TODO
-------------------------
-- Keep testing to ensure the system is stable across Max Resets and daylight system deletions.
-- Make a sanity check for the bitmaps when the rollout opens (in case the scene is re-opened and bitmaps are not found)
-- Move rendering parameters into their own rollout
--
-------------------------
-- TUTORIAL IDEAS
-------------------------
-- 		Ideas for tutorials:
-- 		 > Creating a basic UI using the rollout editor, and how to hook into basic UI events.
--        > Using custom attributes on the root node to save UI state.
-- 		 > Hooking into callbacks to circumvent the Daylight System creation problem in MaxScript.
-- 		 > Rendering previews to bitmaps
--        > Controllers as interpolation engines
--
--================================

( -- open a local context to avoid cluttering the listener.
	
	global SE_Settings_Attr = attributes "SimpleEnvironmentSettings" attribID:#(0x27350c10, 0x23601762)
	(
		-- An attribute which can be added to the root node to store the Simple Environment Editor's settings, so that
		-- it can be re-opened with the same settings after the UI has closed, or after the scene has been saved and re-opened.
		--
		-- All parameter types are listed at: http://help.autodesk.com/view/3DSMAX/2015/ENU/?guid=__files_GUID_461915FA_31A2_49CE_84AF_2544B782ACA3_htm
		--
		-- Note that you can call genClassID in the MaxScript listenre to obtain a valid and unique value for attribID.
		parameters SE_Parameters
		(
			currentBitmapDirectory type:#string
			
			horizon type:#float default:0.5
			env_haze type:#boolean default:true
			env_spherical type:#boolean default:true
			
			bitmap_textureMap type:#textureMap
			physical_sky_textureMap type:#textureMap
			output_textureMap type:#textureMap
			haze_textureMap type:#textureMap
			
			daylight_system type:#string default:"<no daylight system>" -- the name of the daylight system
			ds_compass type:#string default: "<no daylight system>" -- the name of the compass
			
			exposure_sync type:#boolean default:true
			gi_sync type:#boolean default:true
			sun_mult_sync type:#boolean default:true
			sky_mult_sync type:#boolean default:true
			sky_tint_sync type:#boolean default:true
			sky_saturation_sync type:#boolean default:true
			output_amount_sync type:#boolean default:true
			
			exposure_ctrl type:#maxObject
			gi_ctrl type:#maxObject
			sun_mult_ctrl type:#maxObject
			sky_mult_ctrl type:#maxObject
			sky_tint_ctrl type:#maxObject
			sky_saturation_ctrl type:#maxObject
			gain_ctrl type:#maxObject
		)
	)
	
	fn rootNodeHasSettings = (
		-- Checks if the root node already has an instance of SE_Settings_Attr attached to it.
		--
		-- Here, we call custAttributes.getDefs rootNode to identify whether or not the root node contains custom attribute definitions. 
		-- The result of getDefs is undefined if no custom attributes exist, or an array of AttributeDef if some exist, ex: <AttributeDef:SimpleEnvironmentSettings>
		-- We can then compare this attribute definition to see if it matches the SE_Settings_Attr class.
		local definitions = custAttributes.getDefs rootNode
		local definitionFound = false
		if definitions != undefined then (
			for def in definitions while definitionFound == false do  (
				if def == (custAttributes.getDefClass SE_Settings_Attr) then definitionFound = true
			)
		)
		definitionFound
	)
	
	fn findObjectWithClassAndName pClass name:undefined = (
		-- A helper function which finds existing an instance of the given class.
		--
		-- If the name is undefined, then look for any instance of the class.
		local collectionResult = for obj in objects collect (
			if classof obj == pClass then (
				-- If the name is undefined, add it to the result.
				if name == undefined then ( obj ) 
				else (
					-- If the name exists, check if the object's name matches this input name.
					if obj.name == name then ( obj ) 
					else (
						dontCollect
					)
				)
			) else (
				-- if the object is not an instance of the class, don't collect it.
				dontCollect
			)
		)
		-- return the first item in the collection result, or undefined if no such item exists.
		if collectionResult.count > 0 then ( collectionResult[1] ) 
		else ( undefined )
	)
	
	fn findDaylightSystem name:undefined = 
	(
		findObjectWithClassAndName DaylightAssemblyHead name:name
	)
	
	fn findCompass name:undefined =
	(
		findObjectWithClassAndName Compass name:name
	)
	
	fn getParentCompass obj = 
	(
		-- Recursively visits the object's parents and returns the parent compass object, if there is one.
		if classof obj == Compass then ( obj )
		else if obj == undefined or obj.parent == undefined then ( undefined )
		else ( getParentCompass obj.parent )
	)
	
	fn getDecimalTime hrs mins secs = 
	(
		-- Converts the given hour/min/sec input into a decimal time compatible with daylight systems.
		hrs + mins / 60.0 + secs / 3600.0 
	)
	
	fn createBezierFloatControllerFromKeys keyArray = (
		-- keyArray is formatted as such: #( #( <time>, <value> ), #( <time>, <value> ), ... )
		-- where each <time>, <value> pair represents a single key.
		local b = bezier_float()
		
		with animate on (
			for key in keyArray do (
				local t = key[1]
				local v = key[2]
				at time t ( b.value = v )
			)
		)
		
		b
	)
	
	fn initializeEnvironmentSettings = (
		-- Initializes the simple environment settings by creating a custom attribute on the root node.
		-- This way, the settings are conserved when the scene is saved.
		--
		-- Note that this function is only called if rootNodeHasSettings() returns false (defined above).
		local settings = undefined
		
		-- Remove the existing attribute.
		local definitions = custAttributes.getDefs rootNode
		if definitions != undefined then (
			for def in definitions do  (
				if def == (custAttributes.getDefClass SE_Settings_Attr) then (
					custAttributes.delete rootNode def
				)
			)
		)
		
		-- Add the attribute.
		custAttributes.add rootNode SE_Settings_Attr
		
		-- Obtain a reference to the SimpleEnvironmentSettings now that the attribute has been added to the root node.
		settings = rootNode.SimpleEnvironmentSettings
		
		-- Set the default bitmap directory in which the user can look for his bitmaps.
		settings.currentBitmapDirectory = (GetDir #maxroot)
		
		local exposureKeys = #(
			-- time | value
			#( 0, 5.5 ), -- 00:00am (midnight)
			#( 1, 5.5 ), -- 1am
			#( 2, 5.5 ), -- 2am
			#( 3, 5.5 ), -- 3am
			#( 4, 5.5 ), -- 4am
			#( (getDecimalTime 4 30 0 ), 6 ),
			#( 5, 8 ), -- 5am
			#( 6, 11 ), -- 6am
			#( 7, 12 ), -- 7am
			#( 8, 13 ), -- 8am
			#( 9, 13 ), -- 9am
			#( 10, 14 ), -- 10am
			#( 11, 14 ), -- 11am
			#( 12, 14 ), -- 12pm (noon)
			#( 13, 14 ), -- 1pm
			#( 14, 14 ), -- 2pm
			#( 15, 14 ), -- 3pm
			#( 16, 14 ), -- 4pm
			#( 17, 13 ), -- 5pm
			#( 18, 12 ), -- 6pm
			#( 19, 10 ), -- 7pm 
			#( (getDecimalTime 19 30 0 ), 7 ),
			#( 20, 5.5 ), -- 8pm
			#( 21, 5.5 ), -- 9pm
			#( 22, 5.5 ), -- 10pm
			#( 23, 5.5 ), -- 11pm
			#( (getDecimalTime 23 59 59 ), 5.5 ) -- 11:59pm
		)
		
		local giKeys = #(
			#( 0, 0 ), -- 00:00am (midnight)
			#( 1, 0 ), -- 1am
			#( 2, 0 ), -- 2am
			#( 3, 0 ), -- 3am
			#( 4, 0.15 ), -- 4am
			#( 5, 0.3 ), -- 5am
			#( 6, 0.5 ), -- 6am
			#( 7, 0.75 ), -- 7am
			#( 8, 1 ), -- 8am
			#( 9, 1 ), -- 9am
			#( 10, 1 ), -- 10am
			#( 11, 1 ), -- 11am
			#( 12, 1 ), -- 12pm (noon)
			#( 13, 1 ), -- 1pm
			#( 14, 1 ), -- 2pm
			#( 15, 1 ), -- 3pm
			#( 16, 1 ), -- 4pm
			#( 17, 1 ), -- 5pm
			#( 18, 0.75 ), -- 6pm
			#( 19, 0.5 ), -- 7pm 
			#( (getDecimalTime 19 30 0 ), 0.3 ),
			#( 20, 0.2 ), -- 8pm
			#( 21, 0.1 ), -- 9pm
			#( 22, 0 ), -- 10pm
			#( 23, 0 ), -- 11pm
			#( (getDecimalTime 23 59 59 ), 0 ) -- 11:59pm
		)
		
		local sunMultKeys = #(
			-- time | value
			#( 0, 0.02 ), -- 00:00am (midnight)
			#( 1, 0.02 ), -- 1am
			#( 2, 0.05 ), -- 2am
			#( 3, 0.1 ), -- 3am
			#( 4, 0.2 ), -- 4am
			#( 5, 0.3 ), -- 5am
			#( 6, 0.4 ), -- 6am
			#( 7, 0.43 ), -- 7am
			#( 8, 0.6 ), -- 8am
			#( 9, 0.6 ), -- 9am
			#( 10, 1.0 ), -- 10am
			#( 11, 1.0 ), -- 11am
			#( 12, 1.0 ), -- 12pm (noon)
			#( 13, 1.0 ), -- 1pm
			#( 14, 1.0 ), -- 2pm
			#( 15, 1.0 ), -- 3pm
			#( 16, 1.0 ), -- 4pm
			#( 17, 0.8 ), -- 5pm
			#( 18, 0.65 ), -- 6pm
			#( 19, 0.45 ), -- 7pm 
			#( (getDecimalTime 19 30 0 ), 0.37 ),
			#( 20, 0.3 ), -- 8pm
			#( 21, 0.2 ), -- 9pm
			#( 22, 0.05 ), -- 10pm
			#( 23, 0.02 ), -- 11pm
			#( (getDecimalTime 23 59 59 ), 0.02 ) -- 11:59pm
		)
		local skyMultKeys = #(
			-- time | value
			#( 0, 0.02 ), -- 00:00am (midnight)
			#( 1, 0.02 ), -- 1am
			#( 2, 0.05 ), -- 2am
			#( 3, 0.1 ), -- 3am
			#( 4, 0.5 ), -- 4am
			#( 5, 0.52 ), -- 5am
			#( 6, 0.55 ), -- 6am
			#( 7, 0.57 ), -- 7am
			#( 8, 0.6 ), -- 8am
			#( 9, 0.6 ), -- 9am
			#( 10, 1.0 ), -- 10am
			#( 11, 1.0 ), -- 11am
			#( 12, 1.0 ), -- 12pm (noon)
			#( 13, 1.0 ), -- 1pm
			#( 14, 1.0 ), -- 2pm
			#( 15, 1.0 ), -- 3pm
			#( 16, 1.0 ), -- 4pm
			#( 17, 0.8 ), -- 5pm
			#( 18, 0.65 ), -- 6pm
			#( 19, 0.5 ), -- 7pm 
			#( (getDecimalTime 19 30 0 ), 0.49 ),
			#( 20, 0.48 ), -- 8pm
			#( 21, 0.35 ), -- 9pm
			#( 22, 0.15 ), -- 10pm
			#( 23, 0.02 ), -- 11pm
			#( (getDecimalTime 23 59 59 ), 0.02 ) -- 11:59pm
		)
		
		local skyTintKeys = #(
			-- time | value
			#( 0, -1.0 ), -- 00:00am (midnight)
			#( 1, -1.0 ), -- 1am
			#( 2, -1.0 ), -- 2am
			#( 3, -1.0 ), -- 3am
			#( 4, -0.5 ), -- 4am
			#( 5, 0.0 ), -- 5am
			#( 6, 0.0 ), -- 6am
			#( 7, 0.0 ), -- 7am
			#( 8, 0.0 ), -- 8am
			#( 9, 0.0 ), -- 9am
			#( 10, 0.0 ), -- 10am
			#( 11, 0.0 ), -- 11am
			#( 12, 0.0 ), -- 12pm (noon)
			#( 13, 0.0 ), -- 1pm
			#( 14, 0.0 ), -- 2pm
			#( 15, 0.0 ), -- 3pm
			#( 16, 0.0 ), -- 4pm
			#( 17, 0.0 ), -- 5pm
			#( 18, 0.0 ), -- 6pm
			#( 19, 0.0 ), -- 7pm 
			#( 20, -0.3 ), -- 8pm
			#( 21, -0.6 ), -- 9pm
			#( 22, -1.0 ), -- 10pm
			#( 23, -1.0 ), -- 11pm
			#( (getDecimalTime 23 59 59 ), -1.0 ) -- 11:59pm
		)
		
		local skySaturationKeys = #(
			-- time | value
			#( 0, 0.3 ), -- 00:00am (midnight)
			#( 1, 0.3 ), -- 1am
			#( 2, 0.3 ), -- 2am
			#( 3, 0.3 ), -- 3am
			#( 4, 0.3 ), -- 4am
			#( 5, 1.0 ), -- 5am
			#( 6, 1.0 ), -- 6am
			#( 7, 1.0 ), -- 7am
			#( 8, 1.0 ), -- 8am
			#( 9, 1.0 ), -- 9am
			#( 10, 1.0 ), -- 10am
			#( 11, 1.0 ), -- 11am
			#( 12, 1.0 ), -- 12pm (noon)
			#( 13, 1.0 ), -- 1pm
			#( 14, 1.0 ), -- 2pm
			#( 15, 1.0 ), -- 3pm
			#( 16, 1.0 ), -- 4pm
			#( 17, 1.0 ), -- 5pm
			#( 18, 1.0 ), -- 6pm
			#( 19, 1.0 ), -- 7pm 
			#( 20, 1.0 ), -- 8pm
			#( 21, 0.3 ), -- 9pm
			#( 22, 0.3 ), -- 10pm
			#( 23, 0.3 ), -- 11pm
			#( (getDecimalTime 23 59 59 ), 0.3 ) -- 11:59pm
		)
		
		local gainKeys = #(
			-- time | value
			#( 0, 0.02 ), -- 00:00am (midnight)
			#( 1, 0.01 ), -- 1am
			#( 2, 0.02 ), -- 2am
			#( 3, 0.05 ), -- 3am
			#( 4, 0.1 ), -- 4am
			#( 5, 0.25 ), -- 5am
			#( 6, 0.55 ), -- 6am
			#( 7, 0.75 ), -- 7am
			#( 8, 0.85 ), -- 8am
			#( 9, 0.9 ), -- 9am
			#( 10, 1.0 ), -- 10am
			#( 11, 1.0 ), -- 11am
			#( 12, 1.0 ), -- 12pm (noon)
			#( 13, 1.0 ), -- 1pm
			#( 14, 1.0 ), -- 2pm
			#( 15, 1.0 ), -- 3pm
			#( 16, 1.0 ), -- 4pm
			#( 17, 0.9 ), -- 5pm
			#( 18, 0.7 ), -- 6pm
			#( 19, 0.55 ), -- 7pm 
			#( (getDecimalTime 19 30 0 ), 0.315 ),
			#( 20, 0.12 ), -- 8pm
			#( 21, 0.08 ), -- 9pm
			#( 22, 0.035 ), -- 10pm
			#( 23, 0.01 ), -- 11pm
			#( (getDecimalTime 23 59 59 ), 0.01 ) -- 11:59pm
		)
		
		settings.exposure_ctrl = createBezierFloatControllerFromKeys exposureKeys
		settings.gi_ctrl = createBezierFloatControllerFromKeys giKeys
		settings.sun_mult_ctrl = createBezierFloatControllerFromKeys sunMultKeys
		settings.sky_mult_ctrl = createBezierFloatControllerFromKeys skyMultKeys
		settings.sky_tint_ctrl = createBezierFloatControllerFromKeys skyTintKeys
		settings.sky_saturation_ctrl = createBezierFloatControllerFromKeys skySaturationKeys
		settings.gain_ctrl = createBezierFloatControllerFromKeys gainKeys
		
		-- Initialize the environment texture map to an undefined value. This will be set by the user explicitly.
		settings.bitmap_textureMap = undefined
		
		-- Initialize the haze texture map, which acts as a multiplier to the environment bitmap.
		settings.haze_textureMap = output()
		settings.haze_textureMap.output.output_amount = 20.0
		settings.haze_textureMap.name = "Simple_Environment_Haze"
		
		-- Initialize the physical sky texture map.
		local mrPhysSkyInstances = getClassInstances mr_Physical_Sky
		if mrPhysSkyInstances.count > 0 then 
		( 
			-- If an instance of mr_Physical_Sky exists in the scene, use it.
			settings.physical_sky_textureMap = mrPhysSkyInstances[1] 
		) else ( 
			-- Otherwise, create a new instance of mr_Physical_Sky.
			settings.physical_sky_textureMap = mr_Physical_Sky() 
		)
		
		-- Ensure the haze map is hooked up to the mr_Physical_Sky's HazeMap channel.
		settings.physical_sky_textureMap.HazeMap = settings.haze_textureMap
		settings.physical_sky_textureMap.InheritFromSky = false
		settings.physical_sky_textureMap.HazeMapOn = false
		
		-- Initialize the final output exture map, which acts as a multiplier on the sky's color.
		settings.output_textureMap = output()
		settings.output_textureMap.map1 = settings.physical_sky_textureMap
		settings.output_textureMap.name = "Simple_Environment_Output"
		environmentMap = settings.output_textureMap
		
		-- Search for an existing daylight system in the scene.
		local ds = findDaylightSystem()
		if ds != undefined then (
			settings.daylight_system = ds.name
			-- Get the daylight system's compass if it's assigned.
			local c = getParentCompass ds
			if( c != undefined ) then (
				settings.ds_compass = c.name
			)
		)
		settings
	)
	
	fn getSimpleEnvironmentSettings = (
		-- Returns the SimpleEnvironmentSettings singleton. 
		-- If no SimpleEnvironmentSettings singleton exists, it lazily instantiates it, and populate it with the correct values.
		if rootNodeHasSettings() == false then initializeEnvironmentSettings()
		rootNode.SimpleEnvironmentSettings
	)
	
	fn getCurrentDaylightSystem = (
		local settings = getSimpleEnvironmentSettings()
		local ds = findDaylightSystem name:settings.daylight_system
		if ds == undefined then (
			settings.daylight_system = "<no daylight system>"
		)
		ds
	)
	
	fn getCurrentCompass = (
		local settings = getSimpleEnvironmentSettings()
		local c = findCompass name:settings.ds_compass
		if c == undefined then (
			settings.ds_compass = "<no daylight system>"
		)
		c
	)
	
	fn setSimpleEnvironmentDaylightSystem ds doCompass:true = (
		
		local settings = getSimpleEnvironmentSettings()
		
		if ds == undefined then 
		(
			settings.daylight_system = "<no daylight system>" --undefined
			settings.ds_compass = "<no daylight system>" --undefined
			
		) else (
			
			settings.daylight_system = ds.name
			
			if doCompass == true then 
			(
				local parentCompass = getParentCompass ds
				if parentCompass != undefined then settings.ds_compass = parentCompass.name
			)
		)
	)
	
	fn setSimpleEnvironmentCompass c = (
		local settings = getSimpleEnvironmentSettings()
		if c != undefined then (
			settings.ds_compass = c.name
		)
	)
	
	fn updateSimpleEnvironmentMaterialGraph = (
		local settings = getSimpleEnvironmentSettings()
		
		settings.haze_textureMap.map1 = settings.bitmap_textureMap
		
		if settings.bitmap_textureMap != undefined then
		(
			settings.bitmap_textureMap.coords.mappingType = 1 -- "environ" mode instead of "texture"
			
			if settings.env_spherical then 
			(
				-- @ bottom, V_Tiling is 2, V_Offset is 0.25
				-- @ 0.5 value, V_Tiling is 1, V_Offset is 0
				-- Also account for settings.physical_sky_textureMap.HorizonHeight
				-- > tiling must decrease when horizon height is more negative.
				-- > offset must decrease when horizon height is more negative.
				settings.bitmap_textureMap.coords.mapping = 0 -- spherical
				local h = 1.0 - settings.horizon
				local t = h * 2.0 + settings.physical_sky_textureMap.HorizonHeight * 0.1
				local o = (1 - 1/t ) * 0.5
				settings.bitmap_textureMap.coords.U_Tiling = -1
				settings.bitmap_textureMap.coords.V_Tiling = t
				settings.bitmap_textureMap.coords.V_Offset = o
			) else (
				settings.bitmap_textureMap.coords.mapping = 3 -- screen
				settings.bitmap_textureMap.coords.U_Tiling = 1
				settings.bitmap_textureMap.coords.V_Tiling = 1.0
				settings.bitmap_textureMap.coords.V_Offset = 0.0
			)
			
			settings.physical_sky_textureMap.HazeMapOn = settings.env_haze
			
			if settings.env_haze then
			(
				settings.output_textureMap.map1 = settings.physical_sky_textureMap
			) else (
				settings.output_textureMap.map1 = settings.bitmap_textureMap
			)
		) else (
			settings.physical_sky_textureMap.HazeMapOn = false
			--settings.physical_sky_textureMap.InheritFromSky = false
			settings.output_textureMap.map1 = settings.physical_sky_textureMap
		)
	)
	
	fn setSimpleEnvironmentBitmap b = (
		local settings = getSimpleEnvironmentSettings()
		settings.bitmap_textureMap = b
		updateSimpleEnvironmentMaterialGraph()
	)
	
	function createScriptController daylightAssembly controller = (
		
		-- Assumes the daylightAssembly is not undefined and that the settings have been correctly instantiated on the root node.
		fs = float_script()
		
		fs.addTarget "daylight_time" daylightAssembly.sun[#solar_time] --[3] -- #solar_time
		fs.addObject "ctrl_obj" controller
		
		fs.script = 
@"
result = 0.0
at time daylight_time (
	result = ctrl_obj.value
)
result
"
		fs.ThrowOnError = false		
		fs
	)
	
	fn scriptControllerHasDaylightTarget ctrl daylightAssembly = (
		((classof ctrl) == float_script) and ((ctrl.getTarget "daylight_time" asObject:true) == daylightAssembly.sun[#solar_time] )
	)
	
	-- A global variable which will contain a reference to our rollout singleton.
	global SE_Rollout
	
	-- Callback function registration and deregistration which will help our rollout stay synchronized with the
	-- daylight systems in the scene.
	fn registerEventCallbacks = (
		callbacks.addScript #nodeCreated "SE_Rollout.nodeCreatedCallback()" id:#SimpleEnvironmentCallbacks
		callbacks.addScript #nodePostDelete "SE_Rollout.nodePostDeleteCallback()" id:#SimpleEnvironmentCallbacks
		callbacks.addScript #nodeNameSet "SE_Rollout.nodeNameSetCallback()" id:#SimpleEnvironmentCallbacks
	)
	fn deregisterEventCallbacks = (
		callbacks.removeScripts id:#SimpleEnvironmentCallbacks
	)
	
	rollout SE_Rollout "Environment Helper" width:280 height:760
	(
		-- The UI definition of the Simple Environment Editor
		fn isDaylightSystem obj = classof obj == DaylightAssemblyHead
		
		-- A string to denote that no daylight system has been set on the rollout.
		local NO_DAYLIGHT_SYSTEM_MESSAGE = "<no daylight system>"
		
		GroupBox grp1 "Environment" pos:[8,8] width:264 height:176
		
		button btn_env "Set Environment Background" pos:[16,24] width:248 height:16
		label lbl81 "Environment Map Horizon Position" pos:[16,48] width:200 height:16
		slider sld_horizon "" pos:[16,68] width:34 height:80 range:[0.02,0.98,0.5] orient:#vertical ticks:2
		bitmap env_map "Bitmap" pos:[56,64] width:208 height:87 enabled:true fileName:""
		
		checkbox chk_env_haze "Checkbox" pos:[56,160] width:16 height:15 checked:true
		label lbl_clouds "Use as Haze" pos:[72,160] width:72 height:16
		
		checkbox chk_spherical_env "" pos:[152,160] width:16 height:16 checked:true
		label lbl_spherical_env "Spherical" pos:[168,160] width:49 height:12 enabled:true
		
		button btn_clear "Clear" pos:[224,160] width:40 height:16
		
		GroupBox grp2 "Daylight System" pos:[8,192] width:264 height:112
		edittext edit_daylight "" pos:[16,216] width:152 height:14 default:"<no daylight system>" readOnly:true bold:true
		button btn_daylight "New" pos:[176,216] width:40 height:16
		pickbutton pick_daylight "Pick" pos:[224,216] width:40 height:16 message:"Pick a Daylight System" filter:isDaylightSystem
		
		label lbl_north "North Dir:" pos:[16,240] width:48 height:16
		spinner spn_north "" pos:[64,240] width:56 height:16 range:[-360,360,0] type:#float
		
		label lbl_sunHeight "Sun Height:" pos:[144,240] width:56 height:16
		spinner spn_sunHeight "" pos:[208,240] width:56 height:16 range:[0,1000,0] type:#float
		
		label lbl_time "Decimal Time of Day:" pos:[16,280] width:104 height:16
		spinner spn_decimal_time "" pos:[120,280] width:48 height:16 range:[-10000,10000,0] type:#float scale:0.001
		label lbl_hours "Hours" pos:[184,264] width:32 height:16
		spinner spn_hours "" pos:[176,280] width:40 height:16 enabled:false range:[0,23,0] type:#integer scale:1
		label lbl_mins "Mins." pos:[232,264] width:32 height:16
		spinner spn_mins "" pos:[224,280] width:40 height:16 enabled:false range:[0,59,0] type:#integer scale:1
-- 		label lbl_secs "Secs." pos:[224,264] width:32 height:16
-- 		spinner spn_secs "" pos:[216,280] width:48 height:16 range:[0,59,0] type:#integer scale:1
		
		label lbl_exposure "Shutter Speed (Hz):" pos:[16,336] width:96 height:16
		spinner spn_exposure "" pos:[120,336] width:48 height:16 range:[1e-005,100000,0] type:#float
		checkbox chk_sync_exposure "Checkbox" pos:[176,336] width:16 height:16 enabled:true checked:true
		
		label lbl_gi "GI Multiplier:" pos:[16,360] width:64 height:16
		spinner spn_gi "" pos:[120,360] width:48 height:16 range:[0,1,0] type:#float
		checkbox chk_sync_gi "Checkbox" pos:[176,360] width:16 height:16 enabled:true checked:true
		
		label lbl_sun_mult "Sun Multiplier:" pos:[16,384] width:72 height:16
		spinner spn_sun_mult "" pos:[120,384] width:48 height:16 range:[0,10,0] scale:0.1
		checkbox chk_sync_sun_mult "Checkbox" pos:[176,384] width:16 height:16 enabled:true checked:true
		
		label lbl_sky_mult "Sky Multiplier:" pos:[16,408] width:72 height:16
		spinner spn_sky_mult "" pos:[120,408] width:48 height:16 range:[0,10,0] scale:0.1
		checkbox chk_sync_sky_mult "Checkbox" pos:[176,408] width:16 height:16 checked:true
		
		label lbl_sky_tint "Sky Blue/Red:" pos:[16,432] width:72 height:16
		spinner spn_sky_tint "" pos:[120,432] width:48 height:16 range:[-1,1,0]
		checkbox chk_sync_sky_tint "Checkbox" pos:[176,432] width:16 height:16 checked:true
		
		label lbl_sky_saturation "Sky Saturation:" pos:[16,456] width:80 height:16
		spinner spn_sky_saturation "" pos:[120,456] width:48 height:16 range:[0,1,1]
		checkbox chk_sync_sky_saturation "Checkbox" pos:[176,456] width:16 height:16 checked:true
		
		label lbl_gain "Final Multiplier:" pos:[16,480] width:72 height:16
		spinner spn_gain "" pos:[120,480] width:48 height:16 range:[0,100,0]
		checkbox chk_sync_gain "Checkbox" pos:[176,480] width:16 height:16 checked:true
		
		label lbl_sync_exposure "Sync to Time" pos:[200,336] width:64 height:16
		label lbl_sync_gi "Sync to Time" pos:[200,360] width:64 height:16
		label lbl_sync_sun_mult "Sync to Time" pos:[200,384] width:64 height:16
		label lbl_sync_sky_mult "Sync to Time" pos:[200,408] width:64 height:16
		label lbl_sync_sky_tint "Sync to Time" pos:[200,432] width:64 height:16
		label lbl_sync_sky_saturation "Sync to Time" pos:[200,456] width:64 height:16
		label lbl_sync_gain "Sync to Time" pos:[200,480] width:64 height:16
		
		GroupBox grp13 "Preview" pos:[8,520] width:264 height:208
		bitmap bmp_render "Bitmap" pos:[16,560] width:248 height:160
		button btn_render "Render" pos:[184,536] width:80 height:16
		GroupBox grp4 "Rendering Parameters" pos:[8,312] width:264 height:200
		spinner spn_world_horizon "" pos:[128,536] width:48 height:16 range:[-10,10,0]
		label lbl20 "World Horizon Offset:" pos:[16,536] width:112 height:16
		button btn_default "(!) Reset to Default" pos:[8,736] width:264 height:16
		
		-- An array containing the environment UI controls which can be enabled or disabled.
		local environmentControls = #(
			--sld_horizon, 
			chk_env_haze, lbl_clouds,
			chk_spherical_env, lbl_spherical_env,
			btn_clear
		)
		fn enableEnvironmentUI state = (
			for ctrl in environmentControls do ctrl.enabled = state
		)
		
		-- An array containing the rendering and daylight system UI controls which can be enabled or disabled.
		local lightingControls = #(
			lbl_north, spn_north,
			lbl_sunHeight, spn_sunHeight,
			lbl_time, spn_decimal_time, 
			lbl_hours, --spn_hours, 
			lbl_mins, --spn_mins, 
			--lbl_secs, spn_secs,
			lbl_exposure, spn_exposure, chk_sync_exposure, 
			lbl_gi, spn_gi, chk_sync_gi,
			lbl_sun_mult, spn_sun_mult, chk_sync_sun_mult,
			lbl_sky_mult, spn_sky_mult, chk_sync_sky_mult, 
			lbl_sky_tint, spn_sky_tint, chk_sync_sky_tint,
			lbl_sky_saturation, spn_sky_saturation, chk_sync_sky_saturation,
			lbl_gain, spn_gain, chk_sync_gain, 
			lbl_sync_exposure, lbl_sync_gi, lbl_sync_sun_mult, lbl_sync_sky_mult, lbl_sync_sky_tint, lbl_sync_sky_saturation, lbl_sync_gain
		)
		fn enableLightingUI state = 
		(
			btn_daylight.enabled = not state
			for ctrl in lightingControls do ctrl.enabled = state
		)
		
		function generateEnvironmentPreviewBitmap originalBitmap = (
			
			local targetBitmap = bitmap env_map.width env_map.height color:black
			
			if originalBitmap == undefined or originalBitmap.height <= 0 or originalBitmap.width <= 0 then (
				-- do nothing.
			) else (
				copy originalBitmap targetBitmap
			)
			
			-- Draw the red horizon line based on the horizon's slider value.
			-- Note: the horizon line is reversed.
			local sliderRange = sld_horizon.range
			local sliderMin = sliderRange[1]
			local sliderMax = sliderRange[2]
			local horizonPercent = 1.0 - sld_horizon.value --1.0 - ( sld_horizon.value - sliderMin ) / ( sliderMax - sliderMin )
			
			local horizonLinePadding = 1
			local y = int( horizonPercent * targetBitmap.height )
			for i = 1 to targetBitmap.width do (
				local x = i
				for j = (y-horizonLinePadding) to (y+horizonLinePadding) do (
					if j >= 0 and j <= targetBitmap.height do (
						setPixels targetBitmap [x, j] #(red)
					)
				)
			)
			
			targetBitmap
		)
		
		local defaultEnvironmentPreviewBitmap = undefined
		fn updateDisplayedBitmap =
		(
			if defaultEnvironmentPreviewBitmap == undefined then (
				-- lazily initialize the default preview bitmap.
				defaultEnvironmentPreviewBitmap = bitmap env_map.width env_map.height color:black
			)
			
			local settings = getSimpleEnvironmentSettings()
			if settings.bitmap_textureMap == undefined then (
				env_map.bitmap = generateEnvironmentPreviewBitmap defaultEnvironmentPreviewBitmap	
			) else (
				env_map.bitmap = generateEnvironmentPreviewBitmap settings.bitmap_textureMap.bitmap
			)
		)
		
		fn syncFromSavedSettings = 
		(
			local settings = getSimpleEnvironmentSettings()
			
			enableEnvironmentUI (settings.bitmap_textureMap != undefined)
			
			if settings.bitmap_textureMap != undefined and settings.bitmap_textureMap.bitmap != undefined and settings.bitmap_textureMap.bitmap.filename != undefined then (
				btn_env.text = settings.bitmap_textureMap.bitmap.filename
			) else (
				btn_env.text = "Set Environment Background"
			)
			
			sld_horizon.value = settings.horizon
			chk_env_haze.state = settings.env_haze
			chk_spherical_env.state = settings.env_spherical
			
			updateDisplayedBitmap()
			
			local ds = getCurrentDaylightSystem()
			enableLightingUI ( ds != undefined )
			edit_daylight.text = settings.daylight_system
			
			if ds == undefined or ds.sun == undefined or ds.sun[#solar_time] == undefined then
			(
				
				local spinnersToNeutralize = #(
 					spn_hours, spn_mins,
					spn_exposure, spn_gi, spn_sun_mult, spn_sky_mult, 
					spn_sky_tint, spn_sky_saturation, spn_gain
				)
				
				for s in spinnersToNeutralize do (
					if (classof s.controller) != bezier_float then (
						s.controller = bezier_float()
					)
					s.value = 0.0
				)
				
				try (
					SceneExposureControl.exposureControl.shutter_speed.controller = bezier_float()
				) catch ()
				
				try (
					settings.physical_sky_textureMap.multiplier.controller = bezier_float()
				) catch ()
				
				try (
					settings.physical_sky_textureMap.redness.controller = bezier_float()
				) catch ()
				
				try (
					settings.physical_sky_textureMap.saturation.controller = bezier_float()
				) catch ()
				
				try (
					settings.output_textureMap.output.output_amount.controller = bezier_float()
				) catch ()
				
				
			) else (
				
				if spn_decimal_time.controller != ds.sun[#solar_time].controller then (
					spn_decimal_time.controller = ds.sun[#solar_time].controller
				)
					
				if (classof spn_hours.controller) != float_script or (not (scriptControllerHasDaylightTarget spn_hours.controller ds)) then (
					local fs = float_script()
					fs.addTarget "daylight_time" ds.sun[#solar_time] 
					fs.script = "int(daylight_time)"
					spn_hours.controller = fs
				)
				
				if (classof spn_mins.controller) != float_script or (not (scriptControllerHasDaylightTarget spn_mins.controller ds))then (
					local fs = float_script()
					fs.addTarget "daylight_time" ds.sun[#solar_time]
					fs.script = "int( (daylight_time - int(daylight_time)) * 60 )"
					spn_mins.controller = fs
				)
				
				-- Synchronize the exposure value
				try (
					chk_sync_exposure.state = settings.exposure_sync
					spn_exposure.enabled = not chk_sync_exposure.state
					
					if SceneExposureControl.exposureControl != undefined and
						SceneExposureControl.exposureControl.exposure_value != undefined and
						SceneExposureControl.exposureControl.shutter_speed != undefined then (
						
						if settings.exposure_sync then (

							if SceneExposureControl.exposureControl.exposure_mode != 1 then (
								SceneExposureControl.exposureControl.exposure_mode = 1
							)
							
							if (classof SceneExposureControl.exposureControl.shutter_speed.controller) != float_script or (not (scriptControllerHasDaylightTarget SceneExposureControl.exposureControl.shutter_speed.controller ds)) then (
								
								SceneExposureControl.exposureControl[#aperture].controller = bezier_float()
								
-- 								local fs = createScriptController ds rootNode.exposure_ctrl
								local fs = float_script()
								fs.addTarget "daylight_time" ds.sun[#solar_time] --[3] -- #solar_time
								fs.addObject "ctrl_obj" rootNode.exposure_ctrl
								fs.addObject "f_stop_ctrl_obj" SceneExposureControl.exposureControl[#aperture].controller
								fs.script = 
@"
result = 0.0
at time daylight_time (
	result = ctrl_obj.value
)
result
1/(( f_stop_ctrl_obj.value * f_stop_ctrl_obj.value ) / (2 ^ result ))
"
								fs.ThrowOnError = false	
								
								SceneExposureControl.exposureControl.shutter_speed.controller = fs
							)
							
						) else (
							
							if (classof SceneExposureControl.exposureControl.shutter_speed.controller) != bezier_float then (
								SceneExposureControl.exposureControl.shutter_speed.controller = bezier_float()
							)
						)
						
						if SceneExposureControl.exposureControl.shutter_speed.controller != undefined then (
							spn_exposure.controller = SceneExposureControl.exposureControl.shutter_speed.controller
						)
						
					)
					

				) catch (
					lbl_exposure.enabled = false
					spn_exposure.enabled = false
					chk_sync_exposure.state = false
					chk_sync_exposure.enabled = false
					lbl_sync_exposure.enabled = false
				)
				
				-- Synchronize the GI multiplier value
				try (
					chk_sync_gi.state = settings.gi_sync
					spn_gi.enabled = not chk_sync_gi.state
					
					if ds.sky != undefined and ds.sky.multiplier != undefined then (
						
						if settings.gi_sync then (
							if (classof ds.sky.multiplier.controller) != float_script or (not (scriptControllerHasDaylightTarget ds.sky.multiplier.controller ds)) then (
								local fs = createScriptController ds rootNode.gi_ctrl
								ds.sky.multiplier.controller = fs
							)
						) else (
							if (classof ds.sky.multiplier.controller) != bezier_float then (
								ds.sky.multiplier.controller = bezier_float()
							)
						)
						
						if ds.sky.multiplier.controller != undefined then (
							spn_gi.controller = ds.sky.multiplier.controller
						)
					)
					
				) catch (
					lbl_gi.enabled = false
					spn_gi.enabled = false
					chk_sync_gi.state = false
					chk_sync_gi.enabled = false
					lbl_sync_gi.enabled = false
				)
				
				-- Synchronize the sun multiplier value
				try (
					chk_sync_sun_mult.state = settings.sun_mult_sync
					spn_sun_mult.enabled = not chk_sync_sun_mult.state
					
					if ds.sun != undefined and ds.sun.skyMult != undefined then (
						
						if settings.sun_mult_sync then (
							if (classof ds.sun.skyMult.controller) != float_script or (not (scriptControllerHasDaylightTarget ds.sun.skyMult.controller ds)) then (
								local fs = createScriptController ds rootNode.sun_mult_ctrl
								ds.sun.skyMult.controller = fs
							)
						) else (
							if (classof ds.sun.skyMult.controller) != bezier_float then (
								ds.sun.skyMult.controller = bezier_float()
							)
						)
						
						if ds.sun.skyMult.controller != undefined then (
							spn_sun_mult.controller = ds.sun.skyMult.controller
						)
					)
					
				) catch (
					lbl_sun_mult.enabled = false
					spn_sun_mult.enabled = false
					chk_sync_sun_mult.state = false
					chk_sync_sun_mult.enabled = false
					lbl_sync_sun_mult.enabled = false
				)
				
				-- Synchronize the sky multiplier value
				try (
					chk_sync_sky_mult.state = settings.sky_mult_sync			
					spn_sky_mult.enabled = not chk_sync_sky_mult.state
					
					if settings.physical_sky_textureMap != undefined and settings.physical_sky_textureMap.multiplier != undefined then (
						
						if settings.sky_mult_sync then (
							if (classof settings.physical_sky_textureMap.multiplier.controller) != float_script or (not (scriptControllerHasDaylightTarget settings.physical_sky_textureMap.multiplier.controller ds)) then (
								local fs = createScriptController ds rootNode.sky_mult_ctrl
								settings.physical_sky_textureMap.multiplier.controller = fs
							)
						) else (
							if (classof settings.physical_sky_textureMap.multiplier.controller) != bezier_float then (
								settings.physical_sky_textureMap.multiplier.controller = bezier_float()
							)
						)
						
						if settings.physical_sky_textureMap.multiplier.controller != undefined then (
							spn_sky_mult.controller = settings.physical_sky_textureMap.multiplier.controller
						)
					)
					
				) catch (
					lbl_sky_mult.enabled = false
					spn_sky_mult.enabled = false
					chk_sync_sky_mult.state = false
					chk_sync_sky_mult.enabled = false
					lbl_sync_sky_mult.enabled = false
				)
				
				-- Synchronize the sky tint value
				try (
					chk_sync_sky_tint.state = settings.sky_tint_sync
					spn_sky_tint.enabled = not chk_sync_sky_tint.state
					
					if settings.physical_sky_textureMap != undefined and settings.physical_sky_textureMap.redness != undefined then (
						
						if settings.sky_tint_sync then (
							if (classof settings.physical_sky_textureMap.redness.controller) != float_script or (not (scriptControllerHasDaylightTarget settings.physical_sky_textureMap.redness.controller ds)) then (
								local fs = createScriptController ds rootNode.sky_tint_ctrl
								settings.physical_sky_textureMap.redness.controller = fs
							)
						) else (
							if (classof settings.physical_sky_textureMap.redness.controller) != bezier_float then (
								settings.physical_sky_textureMap.redness.controller = bezier_float()
							)
						)
						
						if settings.physical_sky_textureMap.redness.controller != undefined then (
							spn_sky_tint.controller = settings.physical_sky_textureMap.redness.controller
						)
					)
				) catch (
					lbl_sky_tint.enabled = false
					spn_sky_tint.enabled = false
					chk_sync_sky_tint.state = false
					chk_sync_sky_tint.enabled = false
					lbl_sync_sky_tint.enabled = false
				)
				
				-- Synchronize the sky saturation value
				try (
					chk_sync_sky_saturation.state = settings.sky_saturation_sync
					spn_sky_saturation.enabled = not chk_sync_sky_saturation.state
					
					if settings.physical_sky_textureMap != undefined and settings.physical_sky_textureMap.saturation != undefined then (
						
						if settings.sky_saturation_sync then (
							if (classof settings.physical_sky_textureMap.saturation.controller) != float_script or (not (scriptControllerHasDaylightTarget settings.physical_sky_textureMap.saturation.controller ds)) then (
								local fs = createScriptController ds rootNode.sky_saturation_ctrl
								settings.physical_sky_textureMap.saturation.controller = fs
							)
						) else (
							if (classof settings.physical_sky_textureMap.saturation.controller) != bezier_float then (
								settings.physical_sky_textureMap.saturation.controller = bezier_float()
							)
						)
						
						if settings.physical_sky_textureMap.saturation.controller != undefined then (
							spn_sky_saturation.controller = settings.physical_sky_textureMap.saturation.controller
						)
					)
				) catch (
					lbl_sky_saturation.enabled = false
					spn_sky_saturation.enabled = false
					chk_sync_sky_saturation.state = false
					chk_sync_sky_saturation.enabled = false
					lbl_sync_sky_saturation.enabled = false
				)
				
				-- Synchronize the final multiplier (gain) value
				try (
					chk_sync_gain.state = settings.output_amount_sync
					spn_gain.enabled = not chk_sync_gain.state
					
					if settings.output_textureMap != undefined and settings.output_textureMap.output != undefined and settings.output_textureMap.output.output_amount != undefined then (
						
						if settings.output_amount_sync then (
							if (classof settings.output_textureMap.output.output_amount.controller) != float_script or (not (scriptControllerHasDaylightTarget settings.output_textureMap.output.output_amount.controller ds)) then (
								local fs = createScriptController ds rootNode.gain_ctrl
								settings.output_textureMap.output.output_amount.controller = fs
							)
						) else (
							if (classof settings.output_textureMap.output.output_amount.controller) != bezier_float then (
								settings.output_textureMap.output.output_amount.controller = bezier_float()
							)
						)
						
						if settings.output_textureMap.output.output_amount.controller != undefined then (
							spn_gain.controller = settings.output_textureMap.output.output_amount.controller
						)
					)
					
				) catch (
					lbl_gain.enabled = false
					spn_gain.enabled = false
					chk_sync_gain.state = false
					chk_sync_gain.enabled = false
					lbl_sync_gain.enabled = false
				)
				
				local c = getCurrentCompass()
				if c == undefined then (
					spn_north.controller = undefined
					spn_north.value = 0.0
				) else (
					spn_north.controller = c.controller.rotation.controller.z_rotation.controller
				)
				
				
				if ds[#Orbital_Scale] != undefined then (
					if ds[#Orbital_Scale].controller == undefined then ds[#Orbital_Scale].controller = bezier_float()
					spn_sunHeight.controller = ds[#Orbital_Scale].controller
				) else (
					spn_sunHeight.controller = undefined
					spn_sunHeight.value = 0.0
				)
				
				
				try (
					if settings.physical_sky_textureMap.HorizonHeight.controller == undefined then settings.physical_sky_textureMap.HorizonHeight.controller = bezier_float()
					spn_world_horizon.controller = settings.physical_sky_textureMap.HorizonHeight.controller
				) catch (
				)
				
			)
		)
		
		fn doUserEnvironmentMapSelection =
		(
			local settings = getSimpleEnvironmentSettings()
			local bitmapFilename = getBitmapOpenFileName caption:"Select Bitmap" filename:(settings.currentBitmapDirectory)
			if bitmapFilename != undefined and doesFileExist bitmapFilename then (
 				settings.currentBitmapDirectory = bitmapFilename
				local b = Bitmaptexture fileName:bitmapFilename
				setSimpleEnvironmentBitmap b
				syncFromSavedSettings()
			)
		)
-- 		
-- 		-- Applies the values of the spinners to the current daylight system.
-- 		fn applyTimeToDaylightSystem =
-- 		(
-- 			if currentDaylightSystem != undefined then
-- 			(
-- 				setTime currentDaylightSystem spn_hours.value spn_mins.value spn_secs.value
-- 				syncAllToTime()
-- 			)
-- 		)
		
		fn nodeCreatedCallback = 
		(
			-- This is the callback which will be triggered after the user presses the "New" button to
			-- create a Daylight system via DaylightSystemFactory2.Create sunClass:mr_Sun skyClass:mr_Sky.
			-- Since the call to DaylightSystemFactory2.Create doesn't return the created daylight system node, 
			-- we'll use the callback system to obtain the newly created daylight system node.
			local nodeCreated = callbacks.notificationParam()
			if classof nodeCreated == DaylightAssemblyHead then 
			(
				setSimpleEnvironmentDaylightSystem nodeCreated doCompass:false
				syncFromSavedSettings()
			)
			
			-- We'll also use this opportunity to set the newly created compass' rotation, since it occasionally
			-- sets itself to (90,0,0).
			if classof nodeCreated == Compass then 
			(
				nodeCreated.rotation.x = 0
				nodeCreated.rotation.y = 0
				nodeCreated.rotation.z = 0
				setSimpleEnvironmentCompass nodeCreated
				syncFromSavedSettings()
			)
		)
		
		fn nodePostDeleteCallback =
		(
			-- Compared to the nodeCreatedCallback, there are no callbacks.notificationParam() on which
			-- to base our search, so we have to check if the rollout's current daylight system has
			-- been deleted, and take appropriate actions if it has been.
			local foundDaylightSystem = findDaylightSystem name:edit_daylight.text
			setSimpleEnvironmentDaylightSystem foundDaylightSystem
			syncFromSavedSettings()
		)
		
		fn nodeNameSetCallback = 
		(
			-- The result of callbacks.notificationParam() has the form: #(name, name, node), where:
			-- the first item is the old name, the second item is the new name, and the third item is 
			-- the node whose name has been set.
			local callbackParams = callbacks.notificationParam()
			
			local settings = getSimpleEnvironmentSettings()
			
			-- Check if the current daylight system matches the node whose name has been set.
			if callbackParams[1] == settings.daylight_system then ( settings.daylight_system = callbackParams[2] )
			
			-- Check if the current compass matches the node whose name has been set.
			if callbackParams[1] == settings.ds_compass then ( settings.ds_compass = callbackParams[2] )
			
			syncFromSavedSettings()
		)
		
		fn resetRollout = (
			
			if (classof SceneExposureControl.exposureControl) != mr_Photographic_Exposure_Control then (
				SceneExposureControl.exposureControl = mr_Photographic_Exposure_Control()
				SceneExposureControl.exposureControl.f_stop = 8
			)
			
			DaylightSimulationUtilities.SetMentalRayAsProductionRenderer()
			syncFromSavedSettings()
			environmentMap = ((getSimpleEnvironmentSettings()).output_textureMap)
		)
		
-- 		on spn_hours changed val do
-- 		(
-- 			ds = getCurrentDaylightSystem()
-- 			if ds != undefined and ds[#solar_time] != undefined then (
-- 				ds[#solar_time].controller.value = getDecimalTime spn_hours.value spn_mins.value 0.0
-- 			)
-- 		)
-- 		on spn_mins changed val do
-- 		(
-- 			ds = getCurrentDaylightSystem()
-- 			if ds != undefined and ds[#solar_time] != undefined then (
-- 				ds[#solar_time].controller.value = getDecimalTime spn_hours.value spn_mins.value 0.0
-- 			)
-- 		)

		on SE_Rollout open do
		(
			registerEventCallbacks()
			resetRollout()
		)
		on SE_Rollout close do
		(
			deregisterEventCallbacks()
		)
		on btn_env pressed do
		(
			doUserEnvironmentMapSelection()
		)
		on sld_horizon changed value do
		(
			local settings = getSimpleEnvironmentSettings()
			settings.horizon = value
			updateSimpleEnvironmentMaterialGraph()
			syncFromSavedSettings()
		)
		on chk_env_haze changed state do
		(
			local settings = getSimpleEnvironmentSettings()
			settings.env_haze = state
			updateSimpleEnvironmentMaterialGraph()
			syncFromSavedSettings()
		)
		on chk_spherical_env changed state do
		(
			local settings = getSimpleEnvironmentSettings()
			settings.env_spherical = state
			updateSimpleEnvironmentMaterialGraph()
			syncFromSavedSettings()
		)
		on btn_clear pressed do
		(
			setSimpleEnvironmentBitmap undefined
			syncFromSavedSettings()
		)
		on btn_daylight pressed do
		(
			-- Set the environment map directly to mr_physical_sky to supress the user warning when
			-- the daylight system is created.
			local settings = getSimpleEnvironmentSettings()
			
			environmentMap = settings.physical_sky_textureMap -- an instance of mr_physical_sky
			
			DaylightSystemFactory2.Create sunClass:mr_Sun skyClass:mr_Sky
			
			environmentMap = settings.output_textureMap
		)
		on pick_daylight picked obj do
		(
			if isValidNode obj then ( setSimpleEnvironmentDaylightSystem obj ) 
			else ( setSimpleEnvironmentDaylightSystem undefined )
			syncFromSavedSettings()
		)
		on chk_sync_exposure changed state do
		(
			local settings = getSimpleEnvironmentSettings()
			settings.exposure_sync = state
			syncFromSavedSettings()
		)
		on chk_sync_gi changed state do
		(
			local settings = getSimpleEnvironmentSettings()
			settings.gi_sync = state
			syncFromSavedSettings()
		)
		on chk_sync_sun_mult changed state do
		(
			local settings = getSimpleEnvironmentSettings()
			settings.sun_mult_sync = state
			syncFromSavedSettings()
		)
		on chk_sync_sky_mult changed state do
		(
			local settings = getSimpleEnvironmentSettings()
			settings.sky_mult_sync = state
			syncFromSavedSettings()
		)
		on chk_sync_sky_tint changed state do
		(
			local settings = getSimpleEnvironmentSettings()
			settings.sky_tint_sync = state
			syncFromSavedSettings()
		)
		on chk_sync_sky_saturation changed state do
		(
			local settings = getSimpleEnvironmentSettings()
			settings.sky_saturation_sync = state
			syncFromSavedSettings()
		)
		on chk_sync_gain changed state do
		(
			local settings = getSimpleEnvironmentSettings()
			settings.output_amount_sync = state
			syncFromSavedSettings()
		)
		on btn_render pressed do
		(
			local oldText = btn_render.text
			btn_render.text = "Rendering..."
			local b = bitmap bmp_render.width bmp_render.height
			render to:b vfb:false
			bmp_render.bitmap = b
			btn_render.text = oldText
		)
		on spn_world_horizon changed val do
		(
			updateSimpleEnvironmentMaterialGraph()
		)
		on btn_default pressed do
		(
			local result = yesNoCancelBox "Do you want to reset the Environment Helper to its default settings?" title:"Reset to Default" beep:false
			if result == #yes then (
				initializeEnvironmentSettings()
				resetRollout()
			)
		)
	)
	
	deregisterEventCallbacks()
	
	global SE_RolloutFloater
	if SE_RolloutFloater != undefined then closeRolloutFloater SE_RolloutFloater
	SE_RolloutFloater = newRolloutFloater "Environment Helper" 292 790
	addRollout SE_Rollout SE_RolloutFloater
	
	Ok
) -- close the local co